/**
 * @file geode_cursor.c
 * @brief Hardware cursor support for Geode GX1 CPU
 **/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <xf86.h>
#include <xf86Cursor.h>
#include <cursorstr.h>

/* X provides us with big(?) endian data for the cursor shape */
#include <byteswap.h>

#include "geode-fb.h"

/**
 * @brief Hide the cursor
 * @param[in] pScreenInfo FIXME
 **/
static void geode_hide_cursor(ScrnInfoRec *pScreenInfo)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	WRITE_REG32(fPtr->dc_io,DC_UNLOCK, DC_UNLOCK_CODE);
	WRITE_REG32(fPtr->dc_io,DC_GENERAL_CFG,READ_REG32(fPtr->dc_io,DC_GENERAL_CFG) & ~(DC_GCFG_CURE));
}

/**
 * @brief Show the cursor
 * @param[in] pScreenInfo FIXME
 **/
static void geode_show_cursor(ScrnInfoRec *pScreenInfo)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );

	WRITE_REG32(fPtr->dc_io,DC_UNLOCK, DC_UNLOCK_CODE);
	WRITE_REG32(fPtr->dc_io,DC_GENERAL_CFG,READ_REG32(fPtr->dc_io,DC_GENERAL_CFG) | DC_GCFG_CURE);
}

/**
 * @brief setup cursor's colours
 * @param[in] pScreen FIXME
 * @param[in] bg Specifies the color value of cursor background color
 * @param[in] fg Specifies the color value of cursor foreground color.
 *
 * Note: DACs are of type 18 bits per colour.
 **/
static void geode_set_cursor_colours(ScrnInfoRec *pScreenInfo, int bg, int fg)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	unsigned int bkcolour=(unsigned int)bg;
	unsigned int fgcolour=(unsigned int)fg;
	unsigned long value;

	WRITE_REG32(fPtr->dc_io,DC_PAL_ADDRESS, 0x100);
	value = ((bkcolour & 0x000000FC) >> 2) |
			((bkcolour & 0x0000FC00) >> (2 + 8 - 6)) |
			((bkcolour & 0x00FC0000) >> (2 + 16 - 12));
	WRITE_REG32(fPtr->dc_io,DC_PAL_DATA, value);
	value = ((fgcolour & 0x000000FC) >> 2) |
			((fgcolour & 0x0000FC00) >> (2 + 8 - 6)) |
			((fgcolour & 0x00FC0000) >> (2 + 16 - 12));
	WRITE_REG32(fPtr->dc_io,DC_PAL_DATA, value);
}

/**
 * @brief Set cursor's coordinates and enable it
 * @param[in] pScreen FIXME
 * @param[in] xpos Specifies the x-cordinates of the cursor
 * @param[in] ypos Specifies the y co-ordinate of the cursor
 *
 * Note: When the cursor should be partially showed above top of
 * screen, the invisible lines must be skipped, as well the
 * vertical "size" must shrink (Y_OFFSET in DC_CURSOR_Y register).
 **/
static void geode_set_cursor_pos(ScrnInfoRec *pScreenInfo, int xpos, int ypos)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	unsigned long cursor;
	unsigned long xoffset = 0;	/* xoffset in cursor image */

	if (fPtr->use_second_cursor)
		cursor=fPtr->cursor_offset_2;
	else
		cursor=fPtr->cursor_offset_1;
	/*
	 * if the pointer is not visible switch it off
	 */
	if ((xpos < -31) || (ypos < -31)) {
		geode_hide_cursor(pScreenInfo);
		return;
	}

	/*
	 * special handling for top and left border
	 */
	if (xpos < 0) {
		xoffset = -xpos;
		xpos = 0;
	}
	if (ypos < 0) {
		fPtr->cursor_yoffset = -ypos;	/* shrink the cursor */
		ypos = 0;
	}
	else
		fPtr->cursor_yoffset = 0;	/* show full size */

	WRITE_REG32(fPtr->dc_io,DC_UNLOCK, DC_UNLOCK_CODE);
	WRITE_REG32(fPtr->dc_io,DC_CURS_ST_OFFSET, cursor + (fPtr->cursor_yoffset<<3));
	WRITE_REG32(fPtr->dc_io,DC_CURSOR_X, (unsigned long) xpos | ((xoffset) << 11));
	WRITE_REG32(fPtr->dc_io,DC_CURSOR_Y, (unsigned long) ypos | ((fPtr->cursor_yoffset) << 11));
	geode_show_cursor(pScreenInfo);
}

/**
 * @brief This function loads the 32x32 cursor pattern.
 * @param[in] pScreen FIXME
 * @param[in] src Specifies cursor data (big endian)
 *
 * The shape and colour is set by AND and XOR masking of
 * arrays of 32 DWORD.
 * X uses a mask and a shape to describe the cursor. This
 * hardware uses 2 bits for each pixel.
 * m-s (m=mask, s=source)
 * 0-0 Cursor colour 1
 * 0-1 Cursor colour 2
 * 1-0 Transparent background
 * 1-1 Inverted background (not used)
 *
 * X will provide us with the correct cursor image as we set
 * the required format when we register our cursor driver
 *
 * We always load the image into the invisible memory and after
 * loading we switch to the new image
 **/
static void geode_load_cursor_image(ScrnInfoRec *pScreenInfo, unsigned char *src)
{
	GeodeDevRec *fPtr = GEODEDEVPTR( pScreenInfo );
	unsigned int i;
	unsigned long *cursor_image;
	unsigned long cursor;
	unsigned long *lsrc=(unsigned long*)src;

	if (fPtr->use_second_cursor)	/* load the invisible cursor */
		cursor_image=fPtr->cursor_1;
	else
		cursor_image=fPtr->cursor_2;

	for (i = 0; i < 64; i++) {
		if (lsrc)
			cursor_image[i] = bswap_32(lsrc[i]);
		else
			cursor_image[i] = 0xFFFF0000;	/* transparent */
	}

	if (fPtr->use_second_cursor) {
		fPtr->use_second_cursor=0;
		cursor=fPtr->cursor_offset_1;
	}
	else {
		fPtr->use_second_cursor=1;
		cursor=fPtr->cursor_offset_2;
	}
	/*
	 * Update the cursor pointer now
	 */
	WRITE_REG32(fPtr->dc_io,DC_UNLOCK, DC_UNLOCK_CODE);
	WRITE_REG32(fPtr->dc_io,DC_CURS_ST_OFFSET,cursor + (fPtr->cursor_yoffset<<3));
}

/**
 * @brief Flag the hardware cursor usability to X
 * @param[in] pScreen FIXME
 * @param[in] pCursor Complete cursor image description
 *
 * Signal if we could use the hardware cursor in this mode
 * It will be called whenever the cursor shape changes or the video
 * mode changes. This is useful for when the hardware cursor cannot
 * be used in interlaced or doublescan modes.
 *
 * @return TRUE when successfull, FALSE elsewhere
 **/
static Bool geode_use_HW_cursor( ScreenRec *pScreen, CursorRec *pCursor)
{
	ScrnInfoRec *pScreenInfo = XF86SCRNINFO(pScreen);

	if (pScreenInfo->currentMode->Flags & V_DBLSCAN)
		return FALSE;

	return TRUE;
}

/**
 * @brief This function extends the X window system with hardware cursor feature
 * @param[in] pScreen FIXME
 *
 * This hardware supports a four coloured cursor. One can be transparent.
 * Maxmimum size of the cursor could be 32 x 32 pixel. To display the cursor
 * a continuous memory of 256 byte in the offscreen area must exist.
 *
 * @return TRUE on success and FALSE on Failure
 **/
int geode_hw_cursor_init(ScreenRec *pScreen)
{
	ScrnInfoRec *pScreenInfo = xf86Screens[pScreen->myNum];
	GeodeDevRec *fPtr = GEODEDEVPTR(pScreenInfo);
	xf86CursorInfoRec *cursorPtr;

	if ((cursorPtr = xf86CreateCursorInfoRec()) == NULL) {
		xf86DrvMsg( pScreenInfo->scrnIndex, X_ERROR,"Cannot get cursor structure\n");
		return FALSE;
	}
	/*
	 * Our hardware feature:
	 * We need the cursor image as (m=mask bit, s=source bit):
	 * mmmmmmmmmmmmmmmm ssssssssssssssss mmmmmmmmmmmmmmmm ssssssssssssssss
	 * ^_left top pixel ^_left top pixel ^_ 16th pixel    ^_ 16th pixel
	 * So we set:
	 * - HARDWARE_CURSOR_BIT_ORDER_MSBFIRST
	 * - HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_16
	 * - HARDWARE_CURSOR_SWAP_SOURCE_AND_MASK
	 *
	 * And we need m=1/s=0 for transparency:
	 * So we set:
	 * - HARDWARE_CURSOR_INVERT_MASK
	 * - HARDWARE_CURSOR_AND_SOURCE_WITH_MASK
	 *
	 * We use two buffer for the cursor image so we can
	 * update the image without prior hiding the cursor
	 * - HARDWARE_CURSOR_UPDATE_UNHIDDEN
	 */
	cursorPtr->Flags = HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |
			HARDWARE_CURSOR_TRUECOLOR_AT_8BPP |
			HARDWARE_CURSOR_UPDATE_UNHIDDEN |
			HARDWARE_CURSOR_SWAP_SOURCE_AND_MASK |
			HARDWARE_CURSOR_AND_SOURCE_WITH_MASK |
			HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_16 |
			HARDWARE_CURSOR_INVERT_MASK;

	cursorPtr->MaxWidth = 32;
	cursorPtr->MaxHeight = 32;

	cursorPtr->SetCursorColors = geode_set_cursor_colours;
	cursorPtr->SetCursorPosition = geode_set_cursor_pos;
	cursorPtr->LoadCursorImage = geode_load_cursor_image;
	cursorPtr->HideCursor = geode_hide_cursor;
	cursorPtr->ShowCursor = geode_show_cursor;
	cursorPtr->RealizeCursor = NULL; /* ??? */
	cursorPtr->UseHWCursor = geode_use_HW_cursor;

	fPtr->cursorPtr=cursorPtr;

	fPtr->use_second_cursor=0;

	return xf86InitCursor(pScreen, cursorPtr);
}

/**
 * @brief Free the cursor on screen close
 * @param[in] pScreenInfo FIXME
 **/
int geode_hw_cursor_exit(ScreenRec *pScreen)
{
	ScrnInfoRec *pScreenInfo = xf86Screens[pScreen->myNum];
	GeodeDevRec *fPtr = GEODEDEVPTR(pScreenInfo);

	geode_hide_cursor(pScreenInfo);
	xf86DestroyCursorInfoRec((xf86CursorInfoRec*)fPtr->cursorPtr);
	fPtr->cursorPtr=NULL;

	return TRUE;
}

/* end of file geode_cursor.c */
