/**
 * @file geode_driver.c
 * @brief Native driver for the Geode GX1 processor and chipset
 * @author Juergen Beisert <jbeisert@netscape.net>
 *
 * This driver bases on Xorg's fbdev
 */

#undef DEBUG

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <xf86.h>
#include <xf86_OSproc.h>

#include "geode-fb.h"

/**
 * @brief Supported "chipsets" by this driver
 *
 * Corrospondends to the <Section "Device">,
 * entry <Driver> in the xorg.conf
 */
static SymTabRec FBDevChipsets[] =
{
	{
		.token	= 0,
		.name	= "fbdev"
	},
	{ .token = -1, .name = NULL }	/* last entry marker */
};

/**
 * @brief Print a driver identifying message only
 * @param[in] flags FIXME
 *
 * This is one of the to be exported functions this driver
 * must support
 **/
static void FBDevIdentify( int flags )
{
	xf86PrintChipsets( GEODEDEV_NAME, "driver for Geode framebuffer", FBDevChipsets );
}

/**
 * @brief driver record contains the functions needed by
 *        the server after loading the driver module.
 *
 * Note: This contains the information required for each
 *       driver before a ScrnInfoRec has been allocated.
 */
_X_EXPORT DriverRec FrameBufferDriver =
{
	.driverVersion	= GEODEDEV_VERSION,
	.driverName	= GEODEDEV_DRIVER_NAME,
	.Identify	= FBDevIdentify,
	.Probe		= FBDevProbe,
	.AvailableOptions = FBDevAvailableOptions,
	.module		= NULL,
	.refCount	= 0,
	.driverFunc	= FBDevDriverFunc
};

/* ------------------ shared library handling ----------------------- */

#ifdef XFree86LOADER

/**
 * @brief register the function to be called, when the module gets loaded
 */
MODULESETUPPROTO( FBDevSetup );

/**
 * @brief Management description of this driver
 **/
static XF86ModuleVersionInfo FBDevVersRec =
{
	.modname	= GEODEDEV_DRIVER_NAME,
	.vendor		= MODULEVENDORSTRING,
	._modinfo1_	= MODINFOSTRING1,	/* constant MODINFOSTRING1/2 to find */
	._modinfo2_	= MODINFOSTRING2,	/* infoarea with a binary editor or sign tool */
	.xf86version	= XORG_VERSION_CURRENT,
	.majorversion	= GEODEDEV_MAJOR_VERSION,
	.minorversion	= GEODEDEV_MINOR_VERSION,
	.patchlevel	= GEODEDEV_PATCH_LEVEL,
	.abiclass	= ABI_CLASS_VIDEODRV,	/* This is a video driver */
	.abiversion	= ABI_VIDEODRV_VERSION,
	.moduleclass	= NULL,
	.checksum	= { 0, 0, 0, 0}
};

/**
 * @brief Module setup is done by this function
 * @param[in] module Pointer to the geode  module
 * @param[in] opts Driver module options.
 * @param[out] errmaj Error major no
 * @param[out] errmin Error minor no.
 *
 * This function sets up the driver in X list and load the
 * module symbols through xf86loader routines..
 * Its called when the module gets loaded
 *
 * @return FALSE on failure, TRUE on success
 */
pointer FBDevSetup(pointer module, pointer opts, int *errmaj, int *errmin)
{
	static Bool setupDone = FALSE;

	if (setupDone != TRUE) {
		setupDone = TRUE;
		xf86AddDriver( &FrameBufferDriver, module,
				HaveDriverFuncs /* seems ignored! */);
		/*
		 * Tell the loader about symbols from other
		 * modules that this module might refer to.
		 */
		geode_gx1_register_symbols();
		return ( pointer ) TRUE;
	}
	if (errmaj)
		*errmaj = LDR_ONCEONLY;

	return NULL;
}

/**
 * @brief This data is accessed by the loader.  The name must be the
 *        module name followed by "ModuleData".
 */
_X_EXPORT XF86ModuleData geodegx1_drvModuleData =
{
	.vers	=	&FBDevVersRec,
	.setup	=	FBDevSetup,
	.teardown =	NULL
};

#endif /* XFree86LOADER */

/* end of file fbdev.c */
